#include "internal.h"

static int _SV_calculate_change_outerloop(SV_simulation *s, double *variables, double t);
static spm_matrix *_SV_calculate_change_A(SV_simulation *s, double *variables);
static int _SV_calculate_change_innerloop(SV_simulation *s, double *variables, spm_matrix *A,
										  double t);

extern void spm_matrix_double_print(spm_matrix *matrix);

int SV_calculate_change(SV_simulation *s, double *variables, double t)
{
	SV_debug_printf(s, "===== calculate-change: %p =====\n", s);
	SV_debug_printf(s, "time = " SV_DOUBLE_FORMAT "\n", t);
	SV_debug_printf(s, "vars = \n");
	SV_debug_print_double_array(s, variables, s->n_variables);
	
	int result;
	
	while (t > s->max_outer_timestep)
	{
		result = _SV_calculate_change_outerloop(s, variables, s->max_outer_timestep);
		if (result!=0)
		{
			SV_debug_printf(s, "----- failed: %d -----\n", result);
			return result;
		}
		
		t -= s->max_outer_timestep;
	}
	
	result = _SV_calculate_change_outerloop(s, variables, t);
	if (result!=0)
	{
		SV_debug_printf(s, "----- failed: %d -----\n", result);
		return result;
	}
	else
	{
		SV_debug_printf(s, "----- done -----\n");
		return 0;
	}
}

int _SV_calculate_change_outerloop(SV_simulation *s, double *variables, double t)
{
	SV_debug_printf(s, "===== calculate-change-outer =====\n", t);
	SV_debug_printf(s, "time = %llX\n", t);
	
	int i, returnval;

	SV_assert_sanity(s);
	SV_assert_sanity(variables);
	if (t==0) return 0;
	
	/* Calculate the matrix that correlates the derivatives of variables to the second derivatives
	of the variables */
	spm_matrix *A = _SV_calculate_change_A(s, variables);
	SV_assert_sanity(A);
	
	SV_debug_printf(s, "A = \n");
	SV_debug_print_matrix(s, A);
	
	/* Change the variables and derivatives */
	while (t>s->max_inner_timestep)
	{
		returnval = _SV_calculate_change_innerloop(s, variables, A, s->max_inner_timestep);
		if (returnval!=0)
		{
			SV_debug_printf(s, "----- failed: %d -----\n", returnval);
			return returnval;
		}
		t -= s->max_inner_timestep;
	}
	returnval = _SV_calculate_change_innerloop(s, variables, A, t);
	if (returnval!=0)
	{
		SV_debug_printf(s, "----- failed: %d -----\n", returnval);
		return returnval;
	}
	
	spm_matrix_free(A);
	
	/* Calculate new values for the variables that do not have derivatives */
	int *knowns = malloc(sizeof(int)*s->n_variables);
	for (i=0; i<s->n_variables; i++)
	{
		knowns[i] = (s->variables[i]->derivative!=NULL);
	}
	returnval = SV_calculate_completion(s, knowns, variables);
	free(knowns);
	if (returnval!=0)
	{
		SV_debug_printf(s, "----- failed: %d -----\n", returnval);
		return returnval;
	}
	
	SV_debug_printf(s, "vars =\n");
	SV_debug_print_double_array(s, variables, s->n_variables);
	
	SV_debug_printf(s, "----- done -----\n");
	
	return 0;
}

static spm_matrix *_SV_calculate_change_A(SV_simulation *s, double *variables)
{
	double *vars_and_derivs[2];
	vars_and_derivs[0] = variables;
	vars_and_derivs[1] = NULL; /* Shouldn't be used anyway */
	
	/* N and M are the two matrices passed to Gauss(). That is, N[a,b] = P(a, nV[b]) and
	M[a,b] = P(a, V[b]). Actually, this is not strictly true; the column numbers in the matrices are
	determined by the id of the variable rather than its position in V or nV. */
	
	spm_matrix *M = spm_matrix_double_alloc(s->n_constraints*2, s->n_variables);
	spm_matrix *N = spm_matrix_double_alloc(s->n_constraints*2, s->n_variables*2);
	SV_assert_memory(M && N);
	
	/* Populate M and N */
	int i;
	for (i=0; i<s->n_constraints*2; i++)
	{
		SV_expr *function;
		SV_variable_map *partials;
		
		if (i%2 == 0)
		{
			function = s->constraints[i/2]->expr;
			partials = s->constraints[i/2]->expr_partials;
		}
		else
		{
			function = s->constraints[i/2]->deriv;
			partials = s->constraints[i/2]->deriv_partials;
		}
		
		/* function = f_i */
		/* 'partials' are the partial derivatives of 'f_i' */
		
		SV_variable_map_element *e;
		for (e=partials->first; e; e=e->next)
		{
			/* value = P(i, V_j) or P(i, nV_j) */
			double value = SV_expr_eval(e->value.expr, vars_and_derivs);
			
			SV_variable *key1 = e->key1;
			int key2 = e->key2;
			if (key2 && key1->derivative)
			{
				key2 = 0;
				key1 = key1->derivative;
			}
			
			if (key1->derivative && key2==0)
			{
				/* The variable is a member of V */
				spm_matrix_double_set(M, i, key1->id, -value);
			}
			else
			{
				/* The variable is not a member of V because it has no derivative. */
				spm_matrix_double_set(N, i, key1->id + key2*s->n_variables, value);
			}
		}
	}

	SV_debug_printf(s, "N =\n");
	SV_debug_print_matrix(s, N);
	SV_debug_printf(s, "M =\n");
	SV_debug_print_matrix(s, M);

	/* Make sure that the solver has enough room to work even if the matrix is underspecified */
	spm_matrix_resize(M, s->n_constraints*2 + s->n_variables*2, s->n_variables);
	spm_matrix_resize(N, s->n_constraints*2 + s->n_variables*2, s->n_variables*2);
	
	/* Perform the Gaussian elimination */
	int result;
	result = spm_gaussian_matrix_underspecified(N, &spm_double_gaussian_info,
												M, &spm_double_gaussianaugmentation_info);
	SV_assert(result==0, "Gaussian elimination failed. This shouldn't happen.");
	/* N has been freed by spm_gaussian_matrix_underspecified */
	
	/* Resize the matrix to cut out what is unnecessary */
	spm_matrix_resize(M, s->n_variables, s->n_variables);
	
	return M;
}

static int _SV_calculate_change_innerloop(SV_simulation *s, double *variables, spm_matrix *A,
										  double t)
{
	SV_debug_printf(s, "===== calculate-change-inner =====\n");
	SV_debug_printf(s, "t = " SV_DOUBLE_FORMAT "\n", t);
	
	int i;
	
	/* Make a duplicate of 'variables' because the original will be modified */
	double *variables_duplicate = malloc(sizeof(double)*s->n_variables);
	SV_assert_memory(variables_duplicate);
	memcpy(variables_duplicate, variables, sizeof(double)*s->n_variables);
	
	/* Update the values of the variables */
	for (i=0; i<s->n_variables; i++)
	{
		SV_variable *var = s->variables[i];
		
		if (var->derivative)
		{
			/* The var's derivative is known; it is a simple process to modify the var */
			variables[var->id] += t * variables_duplicate[var->derivative->id];
		}
		else
		{
			/* The var's derivative is not known explicitly, so we must guess using the linear
			approximation stored in A */
			spm_matrix_element *e;
			for (e=A->rows[var->id]; e; e=e->next_in_row)
			{
				double value = *(double*)SPM_M_VALUE(e);
				SV_variable *deriv = s->variables[e->col]->derivative;
				variables[var->id] += t * value * variables_duplicate[deriv->id];
			}
		}
	}
	
#ifdef SV_DEBUG_LOGGING
	SV_debug_printf(s, "vars =\n");
	SV_debug_print_double_array(s, variables_duplicate, s->n_variables);
	for (i=0; i<s->n_variables; i++)
	{
		variables_duplicate[i] = variables[i] - variables_duplicate[i];
	}
	SV_debug_printf(s, "delta =\n");
	SV_debug_print_double_array(s, variables_duplicate, s->n_variables);
#endif
	
	free(variables_duplicate);
	
	SV_debug_printf(s, "----- done -----\n");
	
	return 0;
}