#include "internal.h"

/* SV_estimate_individual_root finds values for all of the variables and derivatives that
cause a given expression to become 0. 'knowns' is an array which indicates which variables have
fixed values. 'vars_and_derivs' is an array of two arrays; the first array is the fixed values or
approximations for the variables, and the second array is the approximations for their derivatives.
On return, the answers will be where the approximations used to be. */
static int SV_estimate_individual_root(SV_simulation *s, SV_expr *expr, SV_variable_map *partials,
										int *knowns, double **vars_and_derivs,
										int max_iterations, double tolerance)
{
	SV_debug_printf(s, "===== calculate-completion-root =====\n");
	
	int returnval;
	
	double last_deviation;
	int iterations;
	
	SV_variable_map_element *e;
	
	/* Temporary buffers to hold the partial derivatives of the function with respect to each
	variable and derivative. */
	double *partial_derivatives[2];
	partial_derivatives[0] = malloc(sizeof(double)*s->n_variables);
	partial_derivatives[1] = malloc(sizeof(double)*s->n_variables);
	SV_assert_memory(partial_derivatives[0] && partial_derivatives[1]);
	
	for (iterations=0; ; iterations++)
	{
		SV_debug_printf(s, "===== calculate-completion-root-iteration: %d =====\n", iterations);
		SV_debug_printf(s, "vars =\n");
		SV_debug_print_double_array(s, vars_and_derivs[0], s->n_variables);
		SV_debug_printf(s, "derivs =\n");
		SV_debug_print_double_array(s, vars_and_derivs[1], s->n_variables);
		
		/* Check how good our current approximation is */
		double deviation = SV_expr_eval(expr, vars_and_derivs);
		SV_debug_printf(s, "deviation = " SV_DOUBLE_FORMAT "\n", deviation);
		if (fabs(deviation) <= tolerance)
		{
			SV_debug_printf(s, "----- done -----\n");
			/* We are done */
			returnval = 0;
			break;
		}
		
		/* Make sure we haven't run out of iterations */
		if (iterations==max_iterations)
		{
			SV_debug_printf(s, "----- too-many-iterations -----\n");
			returnval = SV_error_single_maxiter;
			break;
		}
		
		/* Make sure our approximation gets better each round */
		if (iterations!=0 && fabs(deviation) > fabs(last_deviation))
		{
			SV_debug_printf(s, "----- diverging -----\n");
			returnval = SV_error_single_divergent;
			break;
		}
		last_deviation = deviation;
		
		/* Calculate the partial derivatives at our current approximation, and count how many are
		nonzero */
		int n_nonzero_derivatives = 0;
		for (e=partials->first; e; e=e->next)
		{
			/* If is_known is 1, then the variable is one of the 'known' vars and its value should
			not change. */
			int is_known =
				(e->key2==0 && knowns[e->key1->id]) ||
				(e->key1->derivative && e->key2==1 && knowns[e->key1->derivative->id]);
			
			if (is_known)
			{
				/* Just set the partial derivative to 0 so the second loop doesn't try to do
				anything with it. */
				partial_derivatives[e->key2][e->key1->id] = 0;
			}
			else
			{
				double partial_derivative = SV_expr_eval(e->value.expr, vars_and_derivs);
				
				if (partial_derivative!=0) n_nonzero_derivatives++;
				
				partial_derivatives[e->key2][e->key1->id] = partial_derivative;
			}
		}
		
		/* Improve our approximation */
		for (e=partials->first; e; e=e->next)
		{
			double partial_derivative = partial_derivatives[e->key2][e->key1->id];
			
			if (partial_derivative==0)
			{
				/* This can mean one of two things:
				
				1. The gradient of the function w.r.t. this variable is flat.
				2. This variable is fixed, and the first loop set the partial derivative to 0 so it
				wouldn't get used.
				
				In either case, we should just skip.
				*/
				continue;
			}
			
			double change = deviation / partial_derivative / n_nonzero_derivatives;
			
			*SV_var_nderiv(vars_and_derivs, e->key1, e->key2) -= change;
		}
		
		SV_debug_printf(s, "----- continuing -----\n");
	}
	
	SV_debug_printf(s, "----- done: %d -----\n", returnval);
	
	free(partial_derivatives[0]);
	free(partial_derivatives[1]);
	
	return returnval;
}

/* A structure to be used as the user-data pointer for the unknown-value-callback */
typedef struct
{
	SV_simulation *simulation;
	double **vars_and_derivs;
} SV_unknown_value_callback_udata;

static int SV_unknown_value_callback(void *_udata, unsigned int var_index, void *output)
{
	/* This function is called if the Gauss-Jordan elimination algorithm supplied by SPM decides
	that one of the variables can have multiple values. 'var_index' is the "floating" variable that
	can have multiple values. The result should be placed in 'output' as a double. */
	
	SV_unknown_value_callback_udata *udata = _udata;
	SV_simulation *s = udata->simulation;
	
	/* Fill in whatever value was used last round */
	int n = var_index / s->n_variables;
	int v_id = var_index % s->n_variables;
	*(double*)output = udata->vars_and_derivs[n][v_id];
	
	SV_debug_printf(s, "===== unknown-value =====\n");
	SV_debug_printf(s, "variable = %d\n", v_id);
	SV_debug_printf(s, "derivativenum = %d\n", n);
	SV_debug_printf(s, "value = " SV_DOUBLE_FORMAT "\n", *(double*)output);
	SV_debug_printf(s, "----- done -----\n");
	
	/* Tell Gauss-Jordan eliminator that it is okay to proceed */
	return 1; 
}

int SV_calculate_completion(SV_simulation *s, int *knowns, double *vars)
{
	SV_debug_printf(s, "===== calculate-completion: %p =====\n", s);
	SV_debug_printf(s, "vars = \n");
	SV_debug_print_double_array(s, vars, s->n_variables);
	SV_debug_printf(s, "knowns = \n");
	SV_debug_print_int_array(s, knowns, s->n_variables);
	
	int returnval;
	int i,j;
	
	int iteration;
	double last_deviation_magnitude;
	
	double *vars_and_derivs[2];
	vars_and_derivs[0] = vars;
	vars_and_derivs[1] = malloc(sizeof(double)*s->n_variables);
	SV_assert_memory(vars_and_derivs[1]);
	
	double *vars_and_derivs_approx[2];
	vars_and_derivs_approx[0] = malloc(sizeof(double)*s->n_variables);
	vars_and_derivs_approx[1] = malloc(sizeof(double)*s->n_variables);
	SV_assert_memory(vars_and_derivs_approx[0] && vars_and_derivs_approx[1]);
	
	spm_matrix *coefficients = NULL;
	double *constants = NULL;
	
	/* Reset the last-constraint-failed indicator */
	s->failed_constraint = NULL;
	
	/* Prepare a SV_unknown_value_callback_udata object to be passed as userdata to
	SV_unknown_value_callback */
	SV_unknown_value_callback_udata unknown_value_callback_udata;
	unknown_value_callback_udata.simulation = s;
	unknown_value_callback_udata.vars_and_derivs = vars_and_derivs;
	
	/* Our initial estimate for the values of the derivatives will be all 0's */
	bzero(vars_and_derivs[1], sizeof(double)*s->n_variables);
	
	for (iteration=0; ; iteration++)
	{
		SV_debug_printf(s, "===== calculate-completion-iteration: %d =====\n", iteration);
		
		/* Check how good our current approximation is */
		double deviation_magnitude = 0;
		int all_constraints_satisfied = 1;
		for (i=0; i<s->n_constraints; i++)
		{
			SV_constraint *c = s->constraints[i];
			
			double deviation = SV_expr_eval(c->expr, vars_and_derivs);
			double deriv_deviation = SV_expr_eval(c->deriv, vars_and_derivs);
			
			if (fabs(deviation) > c->tolerance) all_constraints_satisfied = 0;
			if (fabs(deriv_deviation) > c->tolerance) all_constraints_satisfied = 0;
			deviation_magnitude += deviation*deviation + deriv_deviation*deriv_deviation;
			
			SV_debug_printf(s, "===== calculate-completion-check: %p =====\n", c);
			SV_debug_printf(s, "deviation = " SV_DOUBLE_FORMAT "\n", deviation);
			SV_debug_printf(s, "deriv-deviation = " SV_DOUBLE_FORMAT "\n", deriv_deviation);
			SV_debug_printf(s, "----- done -----\n");
		}
		
		SV_debug_printf(s, "deviation-magnitude = " SV_DOUBLE_FORMAT "\n", deviation_magnitude);
		
		/* Check if we are done */
		if (all_constraints_satisfied)
		{
			SV_debug_printf(s, "----- done -----\n");
			returnval = 0;
			break;
		}
		
		/* Make sure we haven't run out of iterations */
		if (iteration==s->max_iterations)
		{
			SV_debug_printf(s, "----- too-many-iterations -----\n");
			returnval = SV_error_group_maxiter;
			break;
		}
		
		/* Make sure our approximation gets better each round */
		if (iteration>0 && deviation_magnitude > last_deviation_magnitude)
		{
			SV_debug_printf(s, "----- divergent -----\n");
			returnval = SV_error_group_divergent;
			break;
		}
		last_deviation_magnitude = deviation_magnitude;
		
		
		/* Build a matrix of linear approximations. The coefficients of each approximation will go
		in 'coefficients', and the constant will go in 'constants'. There are twice as many rows
		and columns as there are constraints and variables because each constraint's derivative must
		also be included, and same with each variable's derivative.
		
		The constraint numbered #n will go in the slot at 2*n. Its derivative will go in the slot at
		2*n+1. The variable numbered #n will use the index n, and its derivative will use the index
		n+s->n_variables.
		*/
		coefficients = spm_matrix_alloc(s->n_constraints*2, s->n_variables*2,
										&spm_double_value_info);
		constants = malloc(s->n_constraints*2 * sizeof(double));
		
		for (i=0; i<s->n_constraints; i++)
		{
			SV_constraint *c = s->constraints[i];
			
			for (j=0; j<2; j++) /* Repeat for both the expression and its derivative */
			{
				SV_expr         *expr     = (j==0 ? c->expr          : c->deriv);
				SV_variable_map *partials = (j==0 ? c->expr_partials : c->deriv_partials);
				
				SV_debug_printf(s, "===== calculate-completion-constraint: %p: %d =====\n", c, j);
				
				/* Copy our current guess into the approximation arrays */
				memcpy(vars_and_derivs_approx[0], vars_and_derivs[0], 
						sizeof(double)*s->n_variables);
				memcpy(vars_and_derivs_approx[1], vars_and_derivs[1], 
						sizeof(double)*s->n_variables);
				
				/* Find an approximation that satisfies the constraint (or its derivative, if that's
				what we're looking at right now) */
				returnval = SV_estimate_individual_root(s,
														expr,
														partials,
														knowns,
														vars_and_derivs_approx,
														c->max_iterations,
														c->tolerance);
				if (returnval!=0)
				{
					SV_debug_printf(s, "----- failed: %d -----\n", returnval);
					SV_debug_printf(s, "----- failed-root -----\n");
					s->failed_constraint = c;
					goto exit;
				}
				
				/* Put the linear approximation into the (i*2+j)th slot */
				SV_variable_map_element *e;
				constants[i*2+j] = 0;
				for (e=partials->first; e; e=e->next)
				{
					/* If is_known is 1, then the variable is one of the 'known' vars and should not
					become part of the linear approximation. */
					int is_known =
						(e->key2==0 && knowns[e->key1->id]) ||
						(e->key1->derivative && e->key2==1 && knowns[e->key1->derivative->id]);
					
					if (!is_known)
					{
						double partial = SV_expr_eval(e->value.expr, vars_and_derivs_approx);
						
						int var_index;
						if (e->key2==0) var_index = e->key1->id;
						else if (e->key1->derivative) var_index = e->key1->derivative->id;
						else var_index = s->n_variables+e->key1->id;
						spm_matrix_double_set(coefficients, i*2+j, var_index, partial);
						constants[i*2+j] += partial * (*SV_var_nderiv(vars_and_derivs_approx,
																		e->key1, e->key2));
					}
				}
				
				SV_debug_printf(s, "----- done -----\n");
			}
		}
		
		SV_debug_printf(s, "coeffs =\n");
		SV_debug_print_matrix(s, coefficients);
		SV_debug_printf(s, "constants =\n");
		SV_debug_print_double_array(s, constants, s->n_constraints*2);
		
		/* Our system of linear approximations is now complete. However, it may be underspecified - 
		some of the variables might be "flexible", able to take an infinite number of values. In
		order to make sure the solver can still solve the system even in this case, we need to add
		extra empty rows to the system. */
		spm_matrix_resize(coefficients, s->n_constraints*2 + s->n_variables*2, s->n_variables*2);
		constants = realloc(constants, (s->n_constraints*2 + s->n_variables*2)*sizeof(double));
		SV_assert_memory(constants);
		
		/* Now we are ready to solve the system */
		returnval = spm_gaussian_densevector_underspecified(coefficients,
															&spm_double_gaussian_info,
															constants,
															&spm_double_gaussianaugmentation_info,
															SV_unknown_value_callback,
															&unknown_value_callback_udata);
		/* 'coefficients' has been freed by spm_gaussian_densevector_underspecified */
		coefficients = NULL;
		
		SV_assert(returnval==0, "Solving system of linear approximations somehow failed. That "
		"shouldn't happen.");
		
		/* Our answer is now in 'constants'. */
		memcpy(vars_and_derivs[0], constants,                sizeof(double)*s->n_variables);
		memcpy(vars_and_derivs[1], constants+s->n_variables, sizeof(double)*s->n_variables);
		
		free(constants);
		constants = NULL;
		
		SV_debug_printf(s, "vars =\n");
		SV_debug_print_double_array(s, vars_and_derivs[0], s->n_variables);
		SV_debug_printf(s, "derivs =\n");
		SV_debug_print_double_array(s, vars_and_derivs[1], s->n_variables);
		
		SV_debug_printf(s, "----- continuing -----\n");
	}
	
	SV_debug_printf(s, "----- done -----\n");
	
exit:
	free(vars_and_derivs[1]);
	free(vars_and_derivs_approx[0]);
	free(vars_and_derivs_approx[1]);
	
	if (coefficients) spm_matrix_free(coefficients);
	if (constants) free(constants);
	
	return returnval;
}