#include "internal.h"
#include <stdarg.h>

void SV_debug_summarize_simulation(SV_simulation *s)
{
	int i;
	
	SV_debug_printf(s, "===== simulation-summary: %p =====\n", s);
	SV_debug_printf(s, "n-variables = %d\n", s->n_variables);
	SV_debug_printf(s, "n-constraints = %d\n", s->n_constraints);
	SV_debug_printf(s, "max-iterations = %d\n", s->max_iterations);
	SV_debug_printf(s, "max-inner-timestep = " SV_DOUBLE_FORMAT "\n", s->max_inner_timestep);
	SV_debug_printf(s, "max-outer-timestep = " SV_DOUBLE_FORMAT "\n", s->max_outer_timestep);
	
	for (i=0; i<s->n_variables; i++)
	{
		SV_variable *v = s->variables[i];
		
		SV_debug_printf(s, "===== variable-summary: %p =====\n", v);
		SV_debug_printf(s, "derivative = %p\n", v->derivative);
		SV_debug_printf(s, "id = %d\n", v->id);
		SV_debug_printf(s, "----- done -----\n");
	}
	
	for (i=0; i<s->n_constraints; i++)
	{
		SV_constraint *c = s->constraints[i];
		
		SV_debug_printf(s, "===== constraint-summary: %p =====\n", c);
		SV_debug_printf(s, "expr = {\n", c->expr);
		SV_debug_print_expr(s, c->expr);
		SV_debug_printf(s, "}\n");
		SV_debug_printf(s, "deriv = {\n", c->deriv);
		SV_debug_print_expr(s, c->deriv);
		SV_debug_printf(s, "}\n");
		SV_debug_printf(s, "max-iterations = %d\n", c->max_iterations);
		SV_debug_printf(s, "tolerance = " SV_DOUBLE_FORMAT "\n", c->tolerance);
		SV_debug_printf(s, "id = %d\n", c->id);
		SV_debug_printf(s, "----- done -----\n");
	}
	
	SV_debug_printf(s, "----- done -----\n");
}

void SV_debug_printf(SV_simulation *s, const char *format, ...)
{
	if (s->debug_file)
	{
		va_list l;
		va_start(l, format);
		vfprintf(s->debug_file, format, l);
		va_end(l);
	}
}

void SV_debug_print_int_array(SV_simulation *s, int *array, int count)
{
	if (s->debug_file)
	{
		fprintf(s->debug_file, "[");
		int i;
		for (i=0; i<count; i++)
		{
			fprintf(s->debug_file, "%d ", array[i]);
		}
		fprintf(s->debug_file, "]\n");
	}
}

void SV_debug_print_double_array(SV_simulation *s, double *array, int count)
{
	if (s->debug_file)
	{
		fprintf(s->debug_file, "[");
		int i;
		for (i=0; i<count; i++)
		{
			fprintf(s->debug_file, SV_DOUBLE_FORMAT " ", array[i]);
		}
		fprintf(s->debug_file, "]\n");
	}
}

void SV_debug_print_matrix(SV_simulation *s, spm_matrix *m)
{
	if (s->debug_file)
	{
		spm_matrix_double_write(m, s->debug_file);
	}
}

void SV_debug_print_expr(SV_simulation *s, SV_expr *e)
{	
	if (s->debug_file)
	{
		FILE *f = s->debug_file;
		
		switch(e->type)
		{
			case SV_expr_type_variable:
			{
				fprintf(f, "<var %p (%d)>", e->u.variable.var, e->u.variable.var->id);
				break;
			}
			case SV_expr_type_sum:
			{
				fprintf(f, "(");
				SV_debug_print_expr(s, e->u.sum.a);
				fprintf(f, " + ");
				SV_debug_print_expr(s, e->u.sum.b);
				fprintf(f, ")");
				break;
			}
			case SV_expr_type_product:
			{
				fprintf(f, "(");
				SV_debug_print_expr(s, e->u.product.a);
				fprintf(f, "*");
				SV_debug_print_expr(s, e->u.product.b);
				fprintf(f, ")");
				break;
			}
			case SV_expr_type_constant:
			{
				fprintf(f, SV_DOUBLE_FORMAT, e->u.constant.constant);
				break;
			}
			case SV_expr_type_exponential:
			{
				fprintf(f, "e^");
				SV_debug_print_expr(s, e->u.exponential.power);
				break;
			}
			case SV_expr_type_internal1:
			{
				SV_variable *var = e->u.internal1.internal1;
				fprintf(f, "<derivative of var %p (%d)>", var, var->id);
				break;
			}
			case SV_expr_type_internal2:
			{
				fprintf(f, "id(");
				SV_debug_print_expr(s, e->u.internal2.internal1);
				fprintf(f, ")");
				break;
			}
			default:
			{
				fprintf(f, "<BAD EXPR TYPE>");
			}
		}
	}
}

void SV_debug_flush(SV_simulation *sim)
{
	if (sim->debug_file)
	{
		fflush(sim->debug_file);
	}
}