#include "internal.h"

/* SV_calculate_partials_into will calculate the partial derivatives of 'e', multiply them by 'multiplier', and put the results in 'partials'. It will not modify 'e', but it will take ownership of 'multiplier'. */ 
static void SV_calculate_partials_into(SV_expr *e, SV_expr *multiplier, SV_variable_map *partials)
{
	SV_assert_sanity(e);
	SV_assert_sanity(multiplier);
	SV_assert_sanity(partials);
	
	switch(e->type)
	{
		case SV_expr_type_variable:
		case SV_expr_type_internal1:
		{
			/* Expressions of type 'internal1' represent the derivatives of variables. Variables and 
			the derivatives of variables are treated identically except for the value of 'key2' that
			goes into the map. */
			
			int key2 = (e->type==SV_expr_type_variable ? 0 : 1);
			
			SV_variable_map_element *element;
			int already_exists;
			already_exists = SV_variable_map_find_or_create_element(
								partials, e->u.variable.var, key2, &element);
			
			if (already_exists)
			{
				element->value.expr = SV_expr_new_sum(element->value.expr, multiplier);
			}
			else
			{
				element->value.expr = multiplier;
			}
			break;
		}
		
		case SV_expr_type_sum:
		{
			/* d(a+b) = da + db */
			
			/* Can't use SV_expr_shallowcopy here because if one of the recursive subcalls frees
			'multiplier', then 'multiplier2' would become invalid. That's not what we want. */
			SV_expr *multiplier2 = SV_expr_duplicate(multiplier);
			
			SV_calculate_partials_into(e->u.sum.a, multiplier, partials);
			SV_calculate_partials_into(e->u.sum.b, multiplier2, partials);
			break;
		}
		
		case SV_expr_type_product:
		{	
			/* d(a*b) = a*db + b*da */
			SV_expr *a = e->u.product.a;
			SV_expr *b = e->u.product.b;
			SV_expr *a2 = SV_expr_shallowcopy(a);
			SV_expr *b2 = SV_expr_shallowcopy(b);
			
			/* Can't use SV_expr_shallowcopy here because if one of the recursive subcalls freed
			'multiplier', then 'multiplier2' would become invalid. That's not what we want. */
			SV_expr *multiplier2 = SV_expr_duplicate(multiplier);
			
			SV_calculate_partials_into(a, SV_expr_new_product(multiplier, b2), partials);
			SV_calculate_partials_into(b, SV_expr_new_product(multiplier2, a2), partials);
			break;
		}
		
		case SV_expr_type_constant:
		{
			/* d(c) = 0 */
			SV_expr_free(multiplier);
			break;
		}
		
		case SV_expr_type_exponential:
		{
			/* d(e^x) = e^x * dx */
			SV_expr *x = e->u.exponential.power;
			SV_expr *e_to_the_x = SV_expr_shallowcopy(e); 
			SV_calculate_partials_into(x, SV_expr_new_product(multiplier,e_to_the_x), partials);
			break;
		}
		
		case SV_expr_type_internal2:
		{
			SV_calculate_partials_into(e->u.internal2.internal1, multiplier, partials);
			break;
		}
		
		default:
		{
			SV_error("Invalid expression type");
		}
	}
}

SV_variable_map *SV_find_partials(SV_expr *e)
{
	SV_assert_sanity(e);
	
	SV_variable_map *partials = SV_variable_map_new();
	SV_calculate_partials_into(e, SV_expr_new_constant(1), partials);
	return partials;
}

SV_expr *SV_find_derivative(SV_variable_map *partials)
{
	SV_expr *derivative = SV_expr_new_constant(0);
	
	SV_variable_map_element *e;
	for (e=partials->first; e; e=e->next)
	{
		SV_assert(e->key2 == 0, "Trying to take second derivative of an expression");
		
		SV_expr *x = SV_expr_shallowcopy(e->value.expr);
		SV_expr *y = SV_expr_new_internal1(e->key1);
		derivative = SV_expr_new_sum(SV_expr_new_product(x, y), derivative);
	}
	
	return derivative;
}