#include "internal.h"

SV_expr *SV_expr_new_variable(SV_variable *var)
{
	SV_assert_sanity(var);
	
	SV_expr *e = malloc(sizeof(SV_expr));
	SV_assert_memory(e);
	
	e->type = SV_expr_type_variable;
	e->u.variable.var = var;
	
	return e;
}

SV_expr *SV_expr_new_sum(SV_expr *a, SV_expr *b)
{
	SV_assert_sanity(a && b);
	
	SV_expr *e = malloc(sizeof(SV_expr));
	SV_assert_memory(e);

	e->type = SV_expr_type_sum;
	e->u.sum.a = a;
	e->u.sum.b = b;
	
	return e;
}

SV_expr *SV_expr_new_product(SV_expr *a, SV_expr *b)
{
	SV_assert_sanity(a && b);
	
	SV_expr *e = malloc(sizeof(SV_expr));
	SV_assert_memory(e);
	
	e->type = SV_expr_type_product;
	e->u.product.a = a;
	e->u.product.b = b;
	
	return e;
}

SV_expr *SV_expr_new_constant(double constant)
{
	SV_expr *e = malloc(sizeof(SV_expr));
	SV_assert_memory(e);

	e->type = SV_expr_type_constant;
	e->u.constant.constant = constant;
	
	return e;
}

SV_expr *SV_expr_new_exponential(SV_expr *power)
{
	SV_assert_sanity(power);
	
	SV_expr *e = malloc(sizeof(SV_expr));
	SV_assert_memory(e);
	
	e->type = SV_expr_type_exponential;
	e->u.exponential.power = power;
	
	return e;
}

SV_expr *SV_expr_new_internal1(SV_variable *var)
{
	SV_assert_sanity(var);
	
	SV_expr *e = malloc(sizeof(SV_expr));
	SV_assert_memory(e);
	
	e->type = SV_expr_type_internal1;
	e->u.internal1.internal1 = var;
	
	return e;
}

SV_expr *SV_expr_new_internal2(SV_expr *sub)
{
	SV_assert_sanity(sub);
	
	SV_expr *e = malloc(sizeof(SV_expr));
	SV_assert_memory(e);
	
	e->type = SV_expr_type_internal2;
	e->u.internal2.internal1 = sub;
	
	return e;
}

SV_expr *SV_expr_duplicate(SV_expr *e)
{
	SV_assert_sanity(e);
	
	SV_expr *a, *b;
	switch(e->type)
	{
		case SV_expr_type_variable:
		{
			return SV_expr_new_variable(e->u.variable.var);
		}
		case SV_expr_type_sum:
		{
			a = SV_expr_duplicate(e->u.sum.a);
			b = SV_expr_duplicate(e->u.sum.b);
			return SV_expr_new_sum(a, b);
		}
		case SV_expr_type_product:
		{
			a = SV_expr_duplicate(e->u.product.a);
			b = SV_expr_duplicate(e->u.product.b);
			return SV_expr_new_product(a, b);
		}
		case SV_expr_type_constant:
		{
			return SV_expr_new_constant(e->u.constant.constant);
		}
		case SV_expr_type_exponential:
		{
			return SV_expr_new_exponential(SV_expr_duplicate(e->u.exponential.power));
		}
		case SV_expr_type_internal1:
		{
			return SV_expr_new_internal1(e->u.internal1.internal1);
		}
		case SV_expr_type_internal2:
		{
			return SV_expr_new_internal2(e->u.internal2.internal1);
		}
	}
	
	SV_error("Invalid expression type.");
	return NULL;
}

void SV_expr_free(SV_expr *e)
{
	SV_assert_sanity(e);
	
	switch(e->type)
	{
		case SV_expr_type_variable:
		{
			break;
		}
		case SV_expr_type_sum:
		{
			SV_expr_free(e->u.sum.a);
			SV_expr_free(e->u.sum.b);
			break;
		}
		case SV_expr_type_product:
		{
			SV_expr_free(e->u.product.a);
			SV_expr_free(e->u.product.b);
			break;
		}
		case SV_expr_type_constant:
		{
			break;
		}	
		case SV_expr_type_exponential:
		{
			SV_expr_free(e->u.exponential.power);
			break;
		}
		case SV_expr_type_internal1:
		{
			break;
		}
		case SV_expr_type_internal2:
		{
			/* The definition of an expression of type internal2 is that we do not free the 
			subexpression. */
			break;
		}
		default:
		{
			SV_error("Invalid expression type.");
		}
	}
	free(e);
}

double SV_expr_eval(SV_expr *e, double **vars_and_derivs)
{
	SV_assert_sanity(e);
	
	switch(e->type)
	{
		case SV_expr_type_variable:
		{
			SV_assert(e->u.variable.var->simulation, "Expr being evaluated contains a variable " "that is not part of a simulation"); /* Variable must be installed in a simulation. In theory it will be the same simulation the expression is part of, but there's no way to determine that because SV_expr objects do not carry simulation pointers. */
			return vars_and_derivs[0][e->u.variable.var->id];
		}
		case SV_expr_type_sum:
		{
			double a = SV_expr_eval(e->u.sum.a, vars_and_derivs);
			double b = SV_expr_eval(e->u.sum.b, vars_and_derivs);
			return a+b;
		}
		case SV_expr_type_product:
		{
			double a = SV_expr_eval(e->u.product.a, vars_and_derivs);
			double b = SV_expr_eval(e->u.product.b, vars_and_derivs);
			return a*b;
		}
		case SV_expr_type_constant:
		{
			return e->u.constant.constant;
		}
		case SV_expr_type_exponential:
		{
			double p = SV_expr_eval(e->u.exponential.power, vars_and_derivs);
			return exp(p);
		}
		case SV_expr_type_internal1:
		{
			SV_variable *v = e->u.internal1.internal1;
			SV_assert(v->simulation, "Expr being evaluated contains a variable that is not part "
									 "of a simulation");
			return *SV_var_deriv(vars_and_derivs, v);
		}
		case SV_expr_type_internal2:
		{
			return SV_expr_eval(e->u.internal2.internal1, vars_and_derivs);
		}
	}
	
	SV_error("Invalid expression type.");
	return 0;
}