#include "internal.h"

SV_simulation *SV_simulation_new()
{
	SV_simulation *s = malloc(sizeof(SV_simulation));
	SV_assert_memory(s);
	
	s->n_variables = 0;
	s->variables = NULL;
	
	s->n_constraints = 0;
	s->constraints = NULL;
	
	s->max_iterations = SV_DEFAULT_SIMULATION_MAX_ITERATIONS;
	s->max_inner_timestep = SV_DEFAULT_SIMULATION_MAX_INNER_TIMESTEP;
	s->max_outer_timestep = SV_DEFAULT_SIMULATION_MAX_OUTER_TIMESTEP;
	
	s->debug_file = NULL;
	
	s->failed_constraint = NULL;
	
	return s;
}

void SV_simulation_free(SV_simulation *s)
{
	SV_assert_sanity(s);
	
	while (s->n_variables > 0)
	{
		SV_simulation_remove_variable(s, s->variables[s->n_variables-1]);
	}
	while (s->n_constraints > 0)
	{
		SV_simulation_remove_constraint(s, s->constraints[s->n_constraints-1]);
	}
	
	/* Closing debug_file is the responsibility of the library user, so don't close it here. */
	
	free(s->variables);
	free(s->constraints);
	
	free(s);
}

void SV_simulation_add_variable(SV_simulation *s, SV_variable *var)
{
	SV_assert_sanity(var);
	SV_assert(var->simulation == NULL, "A variable cannot belong to two simulations at once.");
	
	s->variables = realloc(s->variables, sizeof(void*)*(s->n_variables+1));
	s->variables[s->n_variables] = var;
	
	var->simulation = s;
	var->id = s->n_variables;
	
	s->n_variables ++;
}

void SV_simulation_add_constraint(SV_simulation *s, SV_constraint *c)
{
	SV_assert_sanity(c);
	SV_assert(c->simulation == NULL, "A constraint cannot belong to two simulations at once, "
	"because it must belong to the same simulation that its vars do, and its vars cannot belong "
	"to two simulations at once.");
	
	s->constraints = realloc(s->constraints, sizeof(void*)*(s->n_constraints+1));
	s->constraints[s->n_constraints] = c;
	
	c->simulation = s;
	c->id = s->n_constraints;
	
	s->n_constraints ++;
}

void SV_simulation_remove_variable(SV_simulation *s, SV_variable *var)
{
	SV_assert_sanity(var);
	SV_assert(var->simulation == s, "Trying to remove a variable from a simulation it is not a "
	"part of.");
	
	/* Shift down variables that were above the one to be removed */
	memcpy(s->variables+var->id,
		   s->variables+var->id+1,
		   sizeof(void*)*(s->n_variables-var->id-1));
	
	/* Update ID numbers of remaining variables */
	int i;
	for (i=var->id; i<s->n_variables-1; i++)
	{
		s->variables[i]->id = i;
	}
	
	var->simulation = NULL;
	var->id = 0;
	
	s->n_variables --;
}

void SV_simulation_remove_constraint(SV_simulation *s, SV_constraint *c)
{
	SV_assert_sanity(c);
	SV_assert(c->simulation == s, "Trying to remove a constraint from a simulation it is not a "
	"part of.");
	
	/* Shift down constraints that were above the one to be removed */
	memcpy(s->constraints+c->id,
		   s->constraints+c->id+1,
		   sizeof(void*)*(s->n_constraints-c->id-1));
	
	/* Update ID numbers of remaining constraints */
	int i;
	for (i=c->id; i<s->n_constraints-1; i++)
	{
		s->constraints[i]->id = i;
	}
	
	c->simulation = NULL;
	c->id = 0;
	
	s->n_constraints --;
}

void SV_simulation_set_property(SV_simulation *s, SV_simulation_property prop, void *value)
{
	SV_assert_sanity(s);
	switch(prop)
	{
		case SV_simulation_property_max_iterations:
		{
			int max_iterations = *(int*)value;
			SV_assert(max_iterations>0, "max_iterations must be at least 1");
			s->max_iterations = max_iterations;
			break;
		}
		case SV_simulation_property_max_outer_timestep:
		{
			double max_outer_timestep = *(double*)value;
			SV_assert(max_outer_timestep>0, "max_outer_timestep must be greater than 0");
			s->max_outer_timestep = max_outer_timestep;
			break;
		}
		case SV_simulation_property_max_inner_timestep:
		{
			double max_inner_timestep = *(double*)value;
			SV_assert(max_inner_timestep>0, "max_inner_timestep must be greater than 0");
			s->max_inner_timestep = max_inner_timestep;
			break;
		}
		case SV_simulation_property_debug_file:
		{
			FILE *debug_file = *(FILE**)value;
			s->debug_file = debug_file;
			break;
		}
		default:
			SV_error("Invalid simulation property.");
	}
}

void SV_simulation_get_property(SV_simulation *s, SV_simulation_property prop, void *value)
{
	SV_assert_sanity(s);
	switch(prop)
	{
		case SV_simulation_property_max_iterations:
			*(int*)value = s->max_iterations;
			break;
		case SV_simulation_property_max_outer_timestep:
			*(double*)value = s->max_outer_timestep;
			break;
		case SV_simulation_property_max_inner_timestep:
			*(double*)value = s->max_inner_timestep;
			break;
		case SV_simulation_property_debug_file:
			*(FILE**)value = s->debug_file;
			break;
		default:
			SV_error("Invalid simulation property.");
	}
}

SV_constraint *SV_simulation_get_failed_constraint(SV_simulation *s)
{
	SV_assert_sanity(s);
	return s->failed_constraint;
}