#include "solver4.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <spm/spm.h>
#include <spm/spm_value_double.h>
#include <spm/spm_gaussian.h>
#include <spm/spm_utils.h>

typedef struct _SV_variable_map SV_variable_map;
typedef struct _SV_variable_map_element SV_variable_map_element;




/* These macros are used for error handling within the solver. */
#define SV_error(message) do {printf("Fatal error in solver:\n%s: %d\n%s\n", __FILE__, __LINE__, (message)); abort();} while(0)
#define SV_assert(cond, message) do {if (!(cond)) SV_error(message);} while(0)
#define SV_assert_sanity(cond) SV_assert(cond, "Sanity check failed: " #cond)
#define SV_assert_memory(cond) SV_assert(cond, "Memory allocation failed: " #cond)




/* Several functions and macros in the solver take a parameter called 'vars_and_derivs'. It is an
array of arrays that holds values for each variable and for its derivative. Normally, the 'n'th
derivative of 'v' with respect to time is stored at vars_and_derivs[n][v->id]. However, if the
derivative of 'v' is specified as another variable, the derivative is stored at
vars_and_derivs[0][v->derivative->id]. At the moment, 'n' is always zero or one. */

/* SV_expr_eval evaluates an expression, given values for its variables and their derivatives. */
double SV_expr_eval(SV_expr *e, double **vars_and_derivs);

/* SV_var_deriv is a macro that expands to a pointer to the slot in 'vars_and_derivs' that holds the 
derivative of the given variable with respect to time. */
#define SV_var_deriv(vars_and_derivs, var) ((var)->derivative ? \
											&(vars_and_derivs)[0][(var)->derivative->id] : \
											&(vars_and_derivs)[1][(var)->id])

/* SV_var_nderiv is a macro that expands to a pointer to the slot in 'vars_and_derivs' that holds
the 'n'-th derivative of the given variable with respect to time. */
#define SV_var_nderiv(vars_and_derivs, var, n) (n==0 ? \
												&(vars_and_derivs)[0][(var)->id] : \
												SV_var_deriv(vars_and_derivs, var))




/* There are two additional types of expression which are only used within the internals of the
solver. They are referred to as 'internal1' and 'internal2'. */

/* An expression of type 'internal1' represents the derivative of a variable. */
SV_expr *SV_expr_new_internal1(SV_variable *var);

/* An expression of type 'internal2' behaves exactly like the expression that it contains, except
that when the expression of type 'internal2' is freed, it will not free its sub-expression. */
SV_expr *SV_expr_new_internal2(SV_expr *sub);
#define SV_expr_shallowcopy(x) SV_expr_new_internal2(x)




struct _SV_constraint
{
	SV_simulation *simulation; /* The simulation that the constraint is part of, if any. */
	int id; /* An identifier, unique to the simulation, that identifies this constraint */
	
	SV_expr *expr; /* The expression that must be equal to 0 */
	SV_variable_map *expr_partials; /* The partial derivatives of the expression */
	SV_expr *deriv; /* The derivative of the expression with respect to time */
	SV_variable_map *deriv_partials; /* The partial derivatives of 'deriv' */
	
	unsigned int max_iterations; /* The maximum number of times the solver should iterate to try to
	solve this constraint before giving up */
	double tolerance; /* The maximum tolerance for error in this constraint. If |expr| is less than
	tolerance, then the solver will consider the constraint satisfied. */
};

#define SV_DEFAULT_CONSTRAINT_TOLERANCE 0.0000001 /* 10^-7 */
#define SV_DEFAULT_CONSTRAINT_MAX_ITERATIONS 10




struct _SV_variable
{
	SV_simulation *simulation; /* The sim that the variable is part of */
	int id; /* An identifier, unique to the simulation, that identifies this variable. */
	SV_variable *derivative; /* Another variable that is the derivative of this one. If this
	variable's derivative is unspecified, then 'derivative' is NULL. */
};




struct _SV_simulation
{
	/* The number of variables in the simulation */
	unsigned int n_variables;
	
	/* An array of pointers to all the variables in the simulation. A variable's position in this
	array is the same as its 'id' member. */
	SV_variable **variables;
	
	/* The number of constraints in the simulation. */
	unsigned int n_constraints;
	
	/* An array of all the constraints in the simulation. A constraint's position in this array is
	the same as its 'id' member. */
	SV_constraint **constraints;
	
	/* The maximum number of total iterations to go through before giving up */
	unsigned int max_iterations;
	
	/* The largest amount to advance 't' by without recalculating the correlation matrix */
	double max_outer_timestep;
	
	/* The largest amount to advance 't' by in one step */
	double max_inner_timestep;
	
	/* Which constraint was the last one to fail. NULL if no constraints failed on the most recent
	call to SV_calculate_completion. */
	SV_constraint *failed_constraint;
	
	/* The file that debug information should be written to, if any. */
	FILE *debug_file;
};

#define SV_DEFAULT_SIMULATION_MAX_ITERATIONS 10
#define SV_DEFAULT_SIMULATION_MAX_OUTER_TIMESTEP 0.10
#define SV_DEFAULT_SIMULATION_MAX_INNER_TIMESTEP 0.01




/* A SV_variable_map object is a data structure that maps pairs of (variable, integer) to SV_expr
objects or to floating-point numbers.

It is used to store the partial derivatives of expressions: the partial derivative of an
expression 'e' with respect to (the 'n'th derivative of a variable 'v' with respect to time) is
stored in the map under (v,n). */
struct _SV_variable_map_element
{
	SV_variable *key1;
	unsigned int key2;
	
	SV_variable_map_element *next;
	
	union
	{
		SV_expr *expr;
		double number;
	} value;
};

struct _SV_variable_map
{
	SV_variable_map_element *first;
};

/* SV_variable_map_new creates a new empty SV_variable_map object */
SV_variable_map *SV_variable_map_new();

/* SV_variable_map_find_or_create_element searches for (key1,key2) in 'v'. If found, then it puts the element in 'e' and returns 1. If not found, it creates a new element with (key1,key2) as its index, puts it in 'e', and returns 0. */
int SV_variable_map_find_or_create_element(SV_variable_map *v, SV_variable *key1, unsigned int key2, SV_variable_map_element **e);

/* SV_variable_map_expr_free frees a variable vector object whose members are expressions. */
void SV_variable_map_expr_free(SV_variable_map *v);

/* SV_variable_map_double_free frees a variable vector object whose members are doubles. */
void SV_variable_map_double_free(SV_variable_map *v);




/* SV_find_partials calculates the partial derivatives of an expression, in the form of a
SV_variable_map. The partial derivatives thus produced may depend on parts of 'e', so they should
not be used after 'e' is freed. */
SV_variable_map *SV_find_partials(SV_expr *e);

/* SV_find_derivative calculates the derivative of an expression with respect to time, given the
partial derivatives of that expression with respect to each variable. The return value may depend on
parts of 'partials', so it should not be used if 'partials' is freed or invalid. */
SV_expr *SV_find_derivative(SV_variable_map *partials);




#ifdef SV_DEBUG_LOGGING

void SV_debug_printf(SV_simulation *sim, const char *format, ...);
void SV_debug_print_int_array(SV_simulation *sim, int *array, int count);
void SV_debug_print_double_array(SV_simulation *sim, double *array, int count);
void SV_debug_print_matrix(SV_simulation *sim, spm_matrix *m);
void SV_debug_print_expr(SV_simulation *sim, SV_expr *e);
void SV_debug_flush(SV_simulation *sim);

#else

#define SV_debug_printf(sim, ...) (void)0
#define SV_debug_print_int_array(sim, array, count) (void)0
#define SV_debug_print_double_array(sim, array, count) (void)0
#define SV_debug_print_matrix(sim, matrix) (void)0
#define SV_debug_print_expr(sim, expr) (void)0
#define SV_debug_flush(sim) (void)0

#endif

#define SV_DOUBLE_FORMAT "%A"