#ifndef SOLVER4_H
#define SOLVER4_H

typedef struct _SV_simulation SV_simulation;
typedef struct _SV_variable SV_variable;
typedef struct _SV_expr SV_expr;
typedef struct _SV_constraint SV_constraint;




/* SV_expr is a structure that represents an expression in terms of variables. */

typedef enum
{
	SV_expr_type_variable,
	SV_expr_type_sum,
	SV_expr_type_product,
	SV_expr_type_constant,
	SV_expr_type_exponential,
	
	/* Used internally */
	SV_expr_type_internal1,
	SV_expr_type_internal2
} SV_expr_type;

struct _SV_expr
{
	SV_expr_type type;
	union
	{
		struct
		{
			SV_variable *var;
		} variable;
		struct
		{
			SV_expr *a;
			SV_expr *b;
		} sum;
		struct
		{
			SV_expr *a;
			SV_expr *b;
		} product;
		struct
		{
			double constant;
		} constant;
		struct
		{
			SV_expr *power;
		} exponential;
		
		/* Used internally */
		struct
		{
			void *internal1;
		} internal1;
		struct
		{
			void *internal1;
		} internal2;
	} u;
};

/* The SV_expr_alloc_* functions create new expressions. The arguments to the function depend on
what kind of expression it is. The arguments will become "part of" the return value, and will be 
freed when SV_expr_free() is called on it. The return value is not guaranteed to be of the type
specified; for example, SV_expr_new_sum() called with two constant expressions may return a constant 
expression instead of a sum expression.*/
SV_expr *SV_expr_new_variable(SV_variable *var);
SV_expr *SV_expr_new_sum(SV_expr *a, SV_expr *b);
SV_expr *SV_expr_new_product(SV_expr *a, SV_expr *b);
SV_expr *SV_expr_new_constant(double constant);
SV_expr *SV_expr_new_exponential(SV_expr *power);

/* SV_expr_duplicate creates a deep-copy of the given expression. */
SV_expr *SV_expr_duplicate(SV_expr *e);

/* SV_expr_free frees the given expression. */
void SV_expr_free(SV_expr *e);




/* SV_variable is a structure that represents a variable. Variables can exist independently of
simulation objects, but they are useless without a simulation to put them in. */

/* SV_variable_new creates a new variable. The variable's derivative is initially unspecified. The
variable is initially not attached to any simulation object. */
SV_variable *SV_variable_new(void);

/* SV_variable_free frees a variable. The variable must not be part of a simulation. If it is,
remove it from the simulation before freeing it. */
void SV_variable_free(SV_variable *var);

/* SV_variable_set_derivative sets the derivative of a variable to another variable. If 'derivative' 
is NULL, then the variable's derivative will be unspecified. */
void SV_variable_set_derivative(SV_variable *var, SV_variable *derivative);

/* SV_variable_get_id returns the 'ID' of a variable, which will be its index into an array of
variables. */
unsigned int SV_variable_get_id(SV_variable *var);





/* A SV_constraint is a structure that represents a constraint on the value of variables. The most
important part of the constraint is a SV_expr object; the constraint is satisfied if this
expression is equal to 0. */

/* Allocates a new constraint object. The supplied expression will be freed when the constraint
is freed. */
SV_constraint *SV_constraint_new(SV_expr *e);

/* Frees a constraint object */
void SV_constraint_free(SV_constraint *c);

/* Replaces the expression of a constraint object with a new one. The old one will be freed, and the
new one will be freed when the constraint is freed. */
void SV_constraint_replace_expr(SV_constraint *c, SV_expr *e);

/* Constraints have 'properties', which are variables that give the solver hints about how to treat
the constraint. Properties are identified by symbolic constants. They can be read and written with SV_constraint_set_property and SV_constraint_get_property. In SV_constraint_get_property, 'value' is a buffer that the solver will write the property into. In SV_constraint_set_property, 'value' is a buffer that the solver will read the property from. */

typedef enum
{
/* A constraint's 'max_iterations' property is the maximum number of iterations that the solver will
perform when trying to find a root for the constraint. If more than 'max_iterations' are necessary, 
the solver will give up and return an error message. 'max_iterations' is of type int. */
	SV_constraint_property_max_iterations,

/* A constraint's 'tolerance' property is how close the value of the constraint's function must be
to 0 for the constraint to be considered satisfied. 'tolerance' must be greater than zero. It is of
type double.*/
	SV_constraint_property_tolerance
} SV_constraint_property;

void SV_constraint_set_property(SV_constraint *c, SV_constraint_property prop, void *value);
void SV_constraint_get_property(SV_constraint *c, SV_constraint_property prop, void *value);




/* SV_simulation is a structure that represents a group of variables and constraints that bind them
together.

Simulations are created, initially empty, using SV_simulation_new(). Variables and constraints can be added and deleted with SV_simulation_[add|remove]_[variable|constraint]. Simulations are 
destroyed with SV_simulation_free(). */

SV_simulation *SV_simulation_new(void);

void SV_simulation_add_variable(SV_simulation *s, SV_variable *var);
void SV_simulation_remove_variable(SV_simulation *s, SV_variable *var);

void SV_simulation_add_constraint(SV_simulation *s, SV_constraint *c);
void SV_simulation_remove_constraint(SV_simulation *s, SV_constraint *c);

void SV_simulation_free(SV_simulation *s);

/* Simulations have 'properties', which are variables that give the solver hints about how to go
about simulating the system. Properties are identified by symbolic constants. They can be read and
written with SV_simulation_set_property and SV_simulation_get_property. In 
SV_simulation_get_property, 'value' is a buffer that the solver will write the property into. In
SV_simulation_set_property, 'value' is a buffer that the solver will read the property from. */

typedef enum
{
/* A simulation's 'max_iterations' property is the maximum number of iterations that the solver will
perform when trying to find a solution. If more than 'max_iterations' are necessary, the simulation 
will give up and return an error message. 'max_iterations' is of type int. */
	SV_simulation_property_max_iterations,

/* A simulation's 'max_outer_timestep' property is the largest amount that the solver will increase
time by before recalculating the correlation matrix. It must be greater than max_inner_timestep.
'max_outer_timestep' is of type double. */
	SV_simulation_property_max_outer_timestep,

/* A simulation's 'max_inner_timestep' property is the largest amount that the solver will increase 
time by at once. If a time-increase of larger than 'max_inner_timestep' is requested, the solver 
will break the increase up into multiple smaller changes. 'max_inner_timestep' is of type double. */
	SV_simulation_property_max_inner_timestep,

/* If solver4 is compiled with SV_DEBUG_LOGGING, then it can produce detailed debug information for
the operations it performs. The simulation's 'debug_file' property is a FILE* object that the
simulation should write debug information to. If it is NULL, no debug information will be produced.
You are responsible for opening and closing this file. */
	SV_simulation_property_debug_file
} SV_simulation_property;

void SV_simulation_set_property(SV_simulation *s, SV_simulation_property prop, void *value);
void SV_simulation_get_property(SV_simulation *s, SV_simulation_property prop, void *value);




/* Given values for some of the variables in the simulation, SV_calculate_completion() calculates 
the values of the remaining variables.

's' is the simulation to perform the calculations on. 'knowns' and 'variables' are arrays in which
each element corresponds to one variable; the index in the array is specified by the 'id' member
of the corresponding SV_variable. 'knowns' is an array of flags that specifies which variables are
known and which are unknown. 'variables' should be an array of doubles; each element of 'variables' 
corresponds to one variable in the solver. If the value in 'knowns' is 1 for a given variable, then 
the corresponding element of 'variables' is the value of that variable. If the value in 'knowns' is 
a 0, then the corresponding element in 'variables' is an initial estimate for the value of that 
variable.

On return, 'variables' will be filled with actual values for the variables. The return value will be 0 on success, and some other value on failure. The meanings of the error codes are as follows:

SV_error_single_divergent: Successive approximations of the roots of a single constraint are getting
	worse, not better.
SV_error_single_maxiter: Iterated too many times when trying to find the roots of a single 
	constraint.
SV_error_group_divergent: Successive approimations of the answer are getting worse, not better.
SV_error_group_maxiter: Iterated too many times when trying to find the answer.

If the error code is SV_error_single_divergent or SV_error_single_maxiter, then you can figure out
which constraint failed with SV_simulation_get_failed_constraint.

*/

int SV_calculate_completion(SV_simulation *s, int *knowns, double *variables);




/* Given values for all of the variables in the simulation at a given time, SV_calculate_change()
calculates the values of the variables at some point in the future.

's' is the simulation to perform the calculations on. 'variables' is an array containing the current
values for all of the variables. 't' is the amount to advance time by.

On return, 'variables' will be filled with new values for the variables. The return value will be 0
on success one of the error codes on failure. The meanings of the error codes are the same as for
SV_calculate_completion. */

int SV_calculate_change(SV_simulation *s, double *variables, double t);




/* SV_calculate_completion and SV_calculate_change can and do sometimes fail. When they do, here are
the error codes. */

#define SV_ok 0
#define SV_error_single_divergent 1
#define SV_error_single_maxiter 2
#define SV_error_group_divergent 3
#define SV_error_group_maxiter 4

/* If the failure is SV_error_single_divergent or SV_error_single_maxiter, then 
SV_simulation_get_failed_constraint will tell you which constraint failed. The return value will be
NULL if no constraints failed on the last time that one of the SV_calculate_* functions was
called. */

SV_constraint *SV_simulation_get_failed_constraint(SV_simulation *s);




/* If SV_DEBUG_LOGGING is defined, SV_debug_summarize_simulation prints detailed information about
the simulation to the simulation's debug file in a semi-human-readable format. If it is not defined, 
then no action is taken. */
void SV_debug_summarize_simulation(SV_simulation *s);

#endif